//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
// Program        : Simple Graph Algorithms
//
// Author         : Richard E. Pattis
//                  Computer Science Department
//                  Carnegie Mellon University
//                  5000 Forbes Avenue
//                  Pittsburgh, PA 15213-3891
//                  e-mail: pattis@cs.cmu.edu
//
// Maintainer     : Author
// 
//
// Description:
//
//   This program is a driver for a variety of "simple" graph algorithms.
// Typically, the user will load a graph and then perform some operation
// on it. Experiment with the graphs in the "graph files" folder.
//
//   In some cases, you can uncomment the System.out.println(...)
// following the "Illustrate" comment, to better understand the 
// operation of the algorithm.
//
// Known Bugs     : None
//
// Future Plans   : None
//
// Program History:
//   11/27/01: R. Pattis - Operational in Java for 15-111
//   11/18/02: R. Pattis - Added Dijkstra Shortest Path + minor improvements
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


import edu.cmu.cs.pattis.cs151xx.Prompt;
import edu.cmu.cs.pattis.cs151xx.TypedBufferReader;
import edu.cmu.cs.pattis.cs151xx.TypedBufferWriter;
import edu.cmu.cs.pattis.cs151xx.IntegerEdgeValue;

import edu.cmu.cs.pattis.cs151xx.Graph;
import edu.cmu.cs.pattis.cs151xx.HashGraph;
import edu.cmu.cs.pattis.cs151xx.orderedCollections.ArrayPriorityQueue;
import edu.cmu.cs.pattis.cs151xx.orderedCollections.ArrayQueue;

import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;



public class Application {

  //All graphs use Integers for their edge values; these methods help
  //  make retrieving/using these values easier
  
  public static int getEdgeValue (Object o)
  {return ((Integer)o).intValue();}

  public static int getEdgeValue (Graph g, String origin, String destination)
  {return getEdgeValue(g.getEdgeValue(origin,destination));}



  //Sources are nodes with in-degree 0
  //AA: O(nodes)
	public static Set getAllSources (Graph g)
	{
	  Set names  = g.getAllNodes();
	  Set answer = new HashSet();
	  
    for (Iterator it=names.iterator(); it.hasNext(); ) {
      String name = (String)it.next();
	    if (g.inDegree(name) == 0)
	      answer.add(name);
	  }
	  
	  return answer;
	}



  //Sinks are nodes with out-degree 0
  //AA: O(nodes)
	public static Set getAllSinks (Graph g)
	{
	  Set names  = g.getAllNodes();
	  Set answer = new HashSet();
	  
    for (Iterator it=names.iterator(); it.hasNext(); ) {
      String name = (String)it.next();
	    if (g.outDegree(name) == 0)
	      answer.add(name);
	  }
	  
	  return answer;
	}



  //Sort a graph representing an ordering that is not trichotomous:
  //  some nodes are not comparable.
  //Works only for acyclic graphs (detects cyclic ones and throws an exception)
  //Remove sources until there are none (removing a source removes various edges
  //  that create other sources).
  //AA: O(nodes^2)
	public static List topologicalSort (Graph g) throws IllegalArgumentException
	{
	  Graph temp   = new HashGraph(g);
	  List  answer = new ArrayList();
	   
	  for (;temp.getNodeCount()>0;) {
	    Set sources = getAllSources(temp);
	    
	    if (sources.size() == 0)
	      throw new IllegalArgumentException("topologicalSort: cyclic");
	      
	    answer.addAll(sources);
	    //Illstrate
	    //System.out.println("Sources = " + sources);
	    
	    for (Iterator it=sources.iterator(); it.hasNext(); )
        temp.removeNode((String)it.next());
      
	    //Illstrate
	    //System.out.println("Graph now = " + temp);
	  }  
	  
	  return answer;
	}



  //Warshall's Algorithm
  //There is an edge o->d in the transitive closure of a graph g iff
  //  if there is a path o to d in g: its edge here is the smallest sum
  //The reason this works and the ordering of the loops (with middle first)
  //  is a bit subtle
  //AA: O(nodes^3)
  //Could also a variant of use Dijkstra's Shorted Path Algorithm N times
	public static Graph transitiveClosure (Graph g)
	{
	  Set names = g.getAllNodes();
	  Graph answer  = new HashGraph(g);
	  
	  for (Iterator n=names.iterator(); n.hasNext(); ) {
	    String middleName = (String)n.next();
	    for (Iterator o=names.iterator(); o.hasNext(); ) {
	      String originName = (String)o.next();
	      for (Iterator d=names.iterator(); d.hasNext(); ) {
	       String destinationName = (String)d.next();
	       
	       //3 different nodes
	       //origin->middle and middle->destination
	       //either no origin->destination or origin->middle->destination shorter than origin->destination
	       if ( originName != destinationName && originName != middleName &&
	            destinationName != middleName &&
	             
	            answer.hasEdge(originName,middleName) &&
	            answer.hasEdge(middleName,destinationName) && 
	            
	            (!answer.hasEdge(originName,destinationName) ||
	             getEdgeValue(answer,originName,middleName) + getEdgeValue(answer,middleName,destinationName)
	              < getEdgeValue(answer,originName,destinationName)
	            )
	           )
	          answer.addEdge(originName,destinationName,new Integer(getEdgeValue(answer,originName,middleName)+  getEdgeValue(answer,middleName,destinationName)));
	     }}}
	         
	  return answer;
	}



  //For each pair of nodes there are either two edges between them or none
  //AA: O(nodes^2)
	public static boolean isWeaklySymmetric (Graph g)
	{
	  Set nodes = g.getAllNodes(); 
	  for (Iterator o=nodes.iterator(); o.hasNext(); ) {
	    String oName = (String)o.next();
	    for (Iterator d=nodes.iterator(); d.hasNext(); ) {
	      String dName = (String)d.next();
	      if (g.hasEdge(oName,dName) != 
	          g.hasEdge(dName,oName))
	        return false;
	    }
	  }  
	  return true;
	}



  //For each pair of nodes there are either two edges with equal
  //  values between them or none
  //AA: O(nodes^2)
	public static boolean isStronglySymmetric (Graph g)
	{
	  Set nodes = g.getAllNodes(); 
	  for (Iterator o=nodes.iterator(); o.hasNext(); ) {
	    String oName = (String)o.next();
	    for (Iterator d=nodes.iterator(); d.hasNext(); ) {
	      String dName = (String)d.next();
	      //One null; one non-null OR one non-null (other must be too) and equal
	      if ( g.hasEdge(oName,dName) != g.hasEdge(dName,oName) ||
             (g.hasEdge(oName,dName)  &&
	            !g.getEdgeValue(oName,dName).equals(g.getEdgeValue(dName,oName))) )
	        return false;
	    }
	  }
	    
	  return true;
	}



  //
  //AA: O(nodes^s)
	public static void makeSymmetric (Graph g)
	{
	  Set nodes = g.getAllNodes(); 
	  for (Iterator o=nodes.iterator(); o.hasNext(); ) {
	    String oName = (String)o.next();
	    for (Iterator d=nodes.iterator(); d.hasNext(); ) {
	      String dName = (String)d.next();
	      if (g.hasEdge(oName,dName))
	        g.addEdge(dName,oName,g.getEdgeValue(oName,dName));
	    }
	  }
	}



  //Used to show final result of spanningTree
  //Just add up all the edges in the graph
  public static int edgeSum(Graph g)
  {
    int sum = 0;
	  for (Iterator e=g.getAllEdges().iterator(); e.hasNext(); ) 
	    sum += ((Integer)((Graph.Edge)e.next()).getValue()).intValue();
	    
	  return sum;
  }
  
  
  //Kruskal's Algorithm
  //Put every node in its own equivalence class (nodes in the same EC map
  //  to the same value)
  //Put every edge in a priority queue: smaller edge values have higher
  //  priority.
  //Pull out each edge; if its origin/destination are in different equivalence
  //  class, add the edge to the answer and merge the equivalence classes
  //The tree must be weakly-symmetric and connected to guarantee a spanning tree
  //AA: O(nodes * max(nodes,Log edges)): ignoring sanity check
  public static Graph spanningTree (Graph g) throws IllegalArgumentException
	{
	  if (!isWeaklySymmetric(g) || components(g).size() != 1)
	    throw new IllegalArgumentException("spanningTree: might not exist");
	    
	  //A spanning tree is a subgraph containing all nodes but only a subset
	  //  of the edges.
	  Graph answer = new HashGraph(g.getEdgeValueIO());
	  
	  //Map each node into its own equivalence class (a number)
    Map equivalenceClasses = new HashMap();
    int nextEC =0;
	  for (Iterator n=g.getAllNodes().iterator(); n.hasNext(); )
      equivalenceClasses.put(n.next(),new Integer(nextEC++));

    //Put all the edges in a priority queue, smallest edge values have highest priority
	  ArrayPriorityQueue pq = new ArrayPriorityQueue(g.getAllEdges(),new Comparator(){
  	     public int compare(Object o1, Object o2)
  	     {
  	       int i1 = ((Integer)((Graph.Edge)o1).getValue()).intValue();
  	       int i2 = ((Integer)((Graph.Edge)o2).getValue()).intValue();
  	       
  	       return -(i1-i2); //Priority for lowest
  	     }
	     }
	  );
	  
	  //Remove edge values from smallest to biggest
	  while (!pq.isEmpty()) {
	    Graph.Edge e = (Graph.Edge)pq.remove();
	    
	    //Illustrate
	    //System.out.println("Extracting " + e);

      //If the node stradles two equivalence classes, add it to the spanning tree
	    if (!equivalenceClasses.get(e.getOrigin()).equals(equivalenceClasses.get(e.getDestination()))) {
	      answer.addEdge(e,e.getValue());
	    //Illustrate
	      //System.out.println("Adding to answer " + e);
      
	      //Merge the two equivalence clsses
	      Integer target1 = (Integer)equivalenceClasses.get(e.getOrigin());
	      Integer target2 = (Integer)equivalenceClasses.get(e.getDestination());
        for (Iterator n=g.getAllNodes().iterator(); n.hasNext(); ) {
          String name =  (String)n.next();
          Integer ec = (Integer)equivalenceClasses.get(name);
			    if (ec.equals(target1) || ec.equals(target2))
			      equivalenceClasses.put(name, new Integer(nextEC));
			  }
			  nextEC++;
	    }
    }

	  return answer;
	}



  //Explain this when there is more time.
	public static Set components (Graph g)
	{
	  Set answer = new HashSet();
	
    Map covered = new HashMap();
    Set nodes = g.getAllNodes();
	  for (Iterator n=nodes.iterator(); n.hasNext(); )
      covered.put(n.next(),new Boolean(false));

    Graph tc = transitiveClosure(g);
    
	  boolean more = true;
	  while (more) {
	    more = false;
	    for (Iterator n=nodes.iterator(); n.hasNext(); ) {
	      String nName = (String)n.next();
	      if (!((Boolean)covered.get(nName)).booleanValue()) {
	        more = true;
	        Set component = new HashSet();
	        component.add(nName);
	        covered.put(nName,new Boolean(true));
	        Set possible = tc.getOutNodes(nName);
	        for (Iterator p=possible.iterator(); p.hasNext(); ) {
	          String pName = (String)p.next();
	          if (!((Boolean)covered.get(pName)).booleanValue()
	              && tc.hasEdge(pName,nName)) {
	            component.add(pName);
	            covered.put(pName, new Boolean(true));
	          }
	        }
	        answer.add(component);
	      }
	    }
	  }
	  return answer;
	}


  
  
  //Compute the transitive closure, then see if we can find two
  //  nodes a,b where there is a path from a to b and b to a.
  //AA: O(nodes^3)
  //Could just do the topological sort (without storing the order)
  //  and return true if successful and false if not:
  //  then, only O(nodes^2)
	public static boolean acyclic (Graph g)
	{
	  Graph tc    = transitiveClosure(g);
	  Set   names = g.getAllNodes();
	  
	  for (Iterator o=names.iterator(); o.hasNext(); ) {
	    String originName = (String)o.next();
	    for (Iterator d=names.iterator(); d.hasNext(); ) {
	      String destinationName = (String)d.next();
	      if ( tc.hasEdge(originName,destinationName) &&
	           tc.hasEdge(destinationName,originName)     )
	        return false;
	    }
	  }
	        
	  return true;
	}



  //Helper for Dijkstra
  public static int getDCost (Map.Entry o)
  {return getDCost( (List)o.getValue() );}
  
  public static int getDCost (List l)
  {return ((Integer)((List)l).get(0)).intValue();}


  //Return all the information for shortest paths from start in graph g.
  //Includes enough information to compute cost and path.
  //Put all nodes in the working graph: start with cost 0; others with cost infinity
  //Take out one node from working (the one with the smallest distance);
  //  then compute the cost to it, plus the cost to any nodes it can reach,
  //  and if that cost is smaller than the current cost stored there, update
  //  the cost and the node computing the cost came from
	public static Map Dijkstra (Graph g, String start)
	{
	  //Each map is of the form: node -> List[distance,previous node]
	  Map working = new HashMap();
	  Map done    = new HashMap();
	  
	  //Build initial working map: 0 for start; infinity for all others
	  for (Iterator i=g.getAllNodes().iterator(); i.hasNext(); ) {
	    String nodeName = (String)i.next();
	    List l = new ArrayList();
	    l.add(new Integer(start.equals(nodeName)? 0 :Integer.MAX_VALUE));
	    l.add(null);
	    working.put(nodeName,l);
	  }
	  
	  //So long as the working map is not empty
	  while(!working.isEmpty()) {
	  
	    //Find its Map.Entry with the smallest distance
	    Map.Entry smallest = null;
	    for (Iterator i=working.entrySet().iterator(); i.hasNext(); ) {
        Map.Entry t = (Map.Entry)i.next();
	      if (smallest==null || getDCost(t) < getDCost(smallest) ) 
	        smallest = t;
	    }
	    
	    //Illustrate
	    //System.out.println("Smallest = " + smallest);

	    //Check all outgoing edges
	    for (Iterator i=g.getOutEdges((String)smallest.getKey()).iterator(); i.hasNext(); ) {
	      Graph.Edge e = (Graph.Edge)i.next();
	      
	      //If we are still working on the destination, the the distance to smallest + the
	      //  distance from smallest to the destination is less than the current distance
	      //  to the destination, update its distnance and previous node (to be this one) 
	      if (working.containsKey(e.getDestination())) {
	        List l2 = (List)working.get(e.getDestination());
	        if ( getDCost(smallest) + ((Integer)(e.getValue())).intValue() <
	             getDCost(l2) ) {
	          l2.set(0,new Integer(getDCost(smallest) + getEdgeValue(e.getValue())));
	          l2.set(1,smallest.getKey());
	          //Illustrateprotected
	          //System.out.println("list of " + e.getDestination() + " set to " + l2);
	        }
        }
	    }
	    
	    //Move smallest from working to done
	    List l = (List)working.remove(smallest.getKey());
	    done.put(smallest.getKey(),l);
	  }
	  
	  //All Map.Entries have been moved from working to done
	  return done;
	}



  ////////////////////////////////////////////////////////////////////////////////
  //
  // Driver for these Methods
  //
  ////////////////////////////////////////////////////////////////////////////////

	private static char menuPrompt ()
  {
    System.out.println("\n\nGraph Operations");
    System.out.println("  c - clear");
    System.out.println("  l - load graph from file");
    System.out.println("  w - write graph to file");
    System.out.println("  d - display graph");
    System.out.println("  e - addEdge");
    System.out.println("  D - Dijkstra Shortest Path");
    System.out.println("  t - transitiveClosure");
    System.out.println("  s - sources and sinks");
    System.out.println("  S - isWeak/StrongSymmetric");
    System.out.println("  M - makeSymmetric");
    System.out.println("  v - spanningTree");
    System.out.println("  a - acyclic");
    System.out.println("  C - components");
    System.out.println("  T - topologicalSort");
    
    return Prompt.forChar("\nEnter Command","clwdDtqesSMvaCT");
  }


	public static void main(String[] args)
	{
    Graph g = new HashGraph(new IntegerEdgeValue());
 
	  
	  //Test priority queue object using any method
	  for (;;) try {

	    char selection = menuPrompt();
	    
      if (selection == 'l')
        g.load(new TypedBufferReader ("  Enter file (containing Graph) name"),' ');

      else if (selection == 'w')
        g.write(new TypedBufferWriter (Prompt.forString("  Enter file (to contain Graph) name")),' ');
      
      else if (selection == 'c')
        g.clear();
 
      
       else if (selection == 'e') {
        String origin      = Prompt.forString("  Enter origin      node");
        String destination = Prompt.forString("  Enter destination node");
        String value       = Prompt.forString("  Enter origin->destination value");
        g.addEdge(origin,destination,g.getEdgeValueIO().readEdgeValue(value));
      }

 
      else if (selection == 'd')
        System.out.println(g);
 
      else if (selection == 'D') {
        String start = Prompt.forString ("  Enter start node");
        Map result = Dijkstra(g,start);
        System.out.println("Dijkstra Costs/Predecessors map = " + result);
        String stop = Prompt.forString (" Enter node to build path to");
        String path = stop;
        for(;;) {
          stop = (String)((List)result.get(stop)).get(1);
          if (stop == null)
            break;
          path = stop + " " + path;
        }
        System.out.println("Path = " + path);
      
      }else if (selection == 't')
        System.out.println(transitiveClosure(g));
 

      else if (selection == 's') {
        System.out.println("Sources = " + getAllSources(g));
        System.out.println("Sinks   = " + getAllSinks(g));
      }  
 

      else if (selection == 'T') {
        System.out.println("Topological Sort = " + topologicalSort(g));
      }  
 

      else if (selection == 'v') {
        Graph st = spanningTree(g);
        System.out.println("Edge sum = " + edgeSum(st));
        System.out.println("Spanning Tree = " + st);
      }  
 
      else if (selection == 'a')
        System.out.println("acyclic = " + acyclic(g));


      else if (selection == 'C')
        System.out.println("Components = " + components(g));


      else if (selection == 'S') {
        System.out.println("isWeaklySymmetric   = " + isWeaklySymmetric(g));
        System.out.println("isStronglySymmetric = " + isStronglySymmetric(g));
      }  
 

      else if (selection == 'M')
        makeSymmetric(g);
 

      else if (selection == 'q')
        break;
      
      else
        System.out.println("\""+selection+"\" is unknown command");

		}catch (Exception e) {
		  System.out.println("Caught Exception");
	    e.printStackTrace();
	  }
  }


}
